/* -[ClockSketch v7.4]----------------------------------------------------------------------------------------
   https://www.instructables.com/ClockSketch-V7-Part-I/
   
   Modified for Lazy Words
   
   
   Arduino UNO/Nano/Pro Mini (AtMega328, 5V, 16 MHz), DS3231 RTC

   December 2023 - Daniel Cikic

   Serial Baud Rates:
   Arduino: 57600
   nodeMCU: 74880
-------------------------------------------------------------------------------------------------------------- */


// comment below to disable serial in-/output and free some RAM
#define DEBUG

// Lazy Words, uncomment only the one to be used
#define LW_ENG  // english
// #define LW_GER  // german

// nodeMCU - uncomment to compile this sketch for nodeMCU 1.0 / ESP8266, make sure to select the proper board
// type inside the IDE! This mode is NOT supported and only experimental!
// #define NODEMCU

// useWiFi - enable WiFi support, WPS setup only! If no WPS support is available on a router check settings
// further down, set useWPS to false and enter ssid/password there --- no wps on lazy words!!
// #define USEWIFI

// useNTP - enable NTPClient, requires NODEMCU and USEWIFI. This will also enforce AUTODST.
// Configure a ntp server further down below!
// #define USENTP

// RTC selection - uncomment the one you're using, comment all others and make sure pin assignemts for
// DS1302 are correct in the parameters section further down!
// #define RTC_DS1302
// #define RTC_DS1307
#define RTC_DS3231

// autoDST - uncomment to enable automatic DST switching, check Time Change Rules below!
// #define AUTODST

// FADING - uncomment to enable fading effects for dots/digits, other parameters further down below
// #define FADING

// autoBrightness - uncomment to enable automatic brightness adjustments by using a photoresistor/LDR
// #define AUTOBRIGHTNESS

// FastForward will speed up things and advance time, this is only for testing purposes!
// Disables AUTODST, USENTP and USERTC.
// #define FASTFORWARD

/* ----------------------------------------------------------------------------------------------------- */


#include <TimeLib.h>                                 // "Time" by Michael Margolis, used in all configs
#include <EEPROM.h>                                  // required for reading/saving settings to eeprom


/* Start RTC config/parameters-------------------------------------------------------------------------- 
   Check pin assignments for DS1302 (SPI), others are I2C (A4/A5 on Arduino by default)                  
   Currently all types are using the "Rtc by Makuna" library                                             */
#ifdef RTC_DS1302
  #include <ThreeWire.h>
  #include <RtcDS1302.h>
  ThreeWire myWire(7, 6, 8);                                     // IO/DAT, SCLK, CE/RST
  RtcDS1302<ThreeWire> Rtc(myWire);
  #define RTCTYPE "DS1302"
  #define USERTC
#endif

#ifdef RTC_DS1307
  #include <Wire.h>
  #include <RtcDS1307.h>
  RtcDS1307<TwoWire> Rtc(Wire);
  #define RTCTYPE "DS1307"
  #define USERTC
#endif

#ifdef RTC_DS3231
  #include <Wire.h>
  #include <RtcDS3231.h>
  RtcDS3231<TwoWire> Rtc(Wire);
  #define RTCTYPE "DS3231"
  #define USERTC
#endif

#if !defined ( USERTC )
  #pragma message "No RTC selected, check definitions on top of the sketch!"
#endif
/* End RTC config/parameters---------------------------------------------------------------------------- */


/* Start WiFi config/parameters------------------------------------------------------------------------- */ 
#ifdef USEWIFI
  const bool useWPS = true;          // set to false to disable WPS and use credentials below - no wps on lazy words!
  const char* wifiSSID = "maWhyFhy";
  const char* wifiPWD = "5up3r1337r0xX0r!";
#endif
/* End WiFi config/parameters--------------------------------------------------------------------------- */


/* Start NTP config/parameters-------------------------------------------------------------------------- 
   Using NTP will enforce autoDST, so check autoDST/time zone settings below!                            */
#ifdef USENTP
  /* I recommend using a local ntp service (many routers offer them), don't spam public ones with dozens 
     of requests a day, get a rtc! ^^                                                                    */
  //#define NTPHOST "europe.pool.ntp.org"
  #define NTPHOST "192.168.2.1"
  #ifndef AUTODST
    #define AUTODST
  #endif
#endif
/* End NTP config/parameters---------------------------------------------------------------------------- */


/* Start autoDST config/parameters ---------------------------------------------------------------------- 
   Comment/uncomment/add TimeChangeRules as needed, only use 2 (tcr1, tcr2), comment out unused ones!     
   Enabling/disabling autoDST will require to set time again, clock will be running in UTC time if autoDST
   is enabled, only display times are adjusted (check serial monitor with DEBUG defined!)                
   This will also add options for setting the date (Year/Month/Day) when setting time on the clock!      */
#ifdef AUTODST
  #include <Timezone.h>                                          // "Timezone" by Jack Christensen
  TimeChangeRule *tcr;
  //-----------------------------------------------
  /* US */
  // TimeChangeRule tcr1 = {"tcr1", First, Sun, Nov, 2, -360};   // utc -6h, valid from first sunday of november at 2am
  // TimeChangeRule tcr2 = {"tcr2", Second, Sun, Mar, 2, -300};  // utc -5h, valid from second sunday of march at 2am
  //-----------------------------------------------
  /* Europe */
  TimeChangeRule tcr1 = {"tcr1", Last, Sun, Oct, 3, 60};         // standard/winter time, valid from last sunday of october at 3am, UTC + 1 hour (+60 minutes) (negative value like -300 for utc -5h)
  TimeChangeRule tcr2 = {"tcr2", Last, Sun, Mar, 2, 120};        // daylight/summer time, valid from last sunday of march at 2am, UTC + 2 hours (+120 minutes)
  //-----------------------------------------------
  Timezone myTimeZone(tcr1, tcr2);
#endif
/* End autoDST config/parameters ----------------------------------------------------------------------- */


/* Start autoBrightness config/parameters -------------------------------------------------------------- */
uint8_t upperLimitLDR = 120;                      // everything above this value will cause max brightness (according to current level) to be used (if it's higher than this)
uint8_t lowerLimitLDR = 80;                       // everything below this value will cause minBrightness to be used
uint8_t minBrightness = 60;                       // anything below this avgLDR value will be ignored
const bool nightMode = false;                     // nightmode true -> if minBrightness is used, colorizeOutput() will use a single color for everything, using HSV
const uint8_t nightColor[2] = { 0, 70 };          // hue 0 = red, fixed brightness of 70, https://github.com/FastLED/FastLED/wiki/FastLED-HSV-Colors
float factorLDR = 1.0;                            // try 0.5 - 2.0, compensation value for avgLDR. Set dbgLDR true & define DEBUG and watch the serial monitor. Looking...
const bool dbgLDR = false;                        // ...for values roughly in the range of 120-160 (medium room light), 40-80 (low light) and 0 - 20 in the dark
#ifdef NODEMCU
  uint8_t pinLDR = 0;                             // LDR connected to A0 (nodeMCU only offers this one)
#else
  uint8_t pinLDR = 1;                             // LDR connected to A1 (in case somebody flashes this sketch on arduino and already has an ldr connected to A1)
#endif
uint8_t intervalLDR = 75;                         // read value from LDR every 75ms (most LDRs have a minimum of about 30ms - 50ms)
uint16_t avgLDR = 0;                              // we will average this value somehow somewhere in readLDR();
uint16_t lastAvgLDR = 0;                          // last average LDR value we got
/* End autoBrightness config/parameters ---------------------------------------------------------------- */


#define SKETCHNAME "ClockSketch v7.4"
#define CLOCKNAME "Lazy Words v1"


/* Start button config/pins----------------------------------------------------------------------------- */
#ifdef NODEMCU
  const uint8_t buttonA = 13;                                    // momentary push button, 1 pin to gnd, 1 pin to d7 / GPIO_13
  const uint8_t buttonB = 14;                                    // momentary push button, 1 pin to gnd, 1 pin to d5 / GPIO_14
#else
  const uint8_t buttonA = 3;                                     // momentary push button, 1 pin to gnd, 1 pin to d3
  const uint8_t buttonB = 4;                                     // momentary push button, 1 pin to gnd, 1 pin to d4
#endif
/* End button config/pins------------------------------------------------------------------------------- */


/* Start basic appearance config------------------------------------------------------------------------ */
uint8_t displayMode = 0;                                         // 0 = 24h mode, 1 = 12h mode ("1" will also override setting that might be written to EEPROM!) -- not on Lazy Words!
uint8_t colorMode = 0;                                           // different color modes, setting this to anything else than zero will overwrite values written to eeprom -- not on Lazy Words!
uint16_t colorSpeed = 600;                                       // controls how fast colors change, smaller = faster (interval in ms at which color moves inside colorizeOutput();)
const bool colorPreview = true;                                  // true = preview selected palette/colorMode using "8" on all positions for 3 seconds
const uint8_t colorPreviewDuration = 3;                          // duration in seconds for previewing palettes/colorModes if colorPreview is enabled/true
const bool reverseColorCycling = true;                           // true = reverse color movements
const uint8_t brightnessLevels[3] {80, 130, 240};                // 0 - 255, brightness Levels (min, med, max) - index (0-2) will be saved to eeprom
uint8_t brightness = brightnessLevels[0];                        // default brightness if none saved to eeprom yet / first run
#ifdef FADING
  uint8_t fadePixels = 2;                                        // fade pixels, 0 = disabled, 1 = only fade out pixels turned off, 2 = fade old out and fade new in
  uint8_t fadeDelay = 30;                                        // milliseconds between each fading step, 5-25 should work okay-ish
#endif
/* End basic appearance config-------------------------------------------------------------------------- */


/* End of basic config/parameters section */


/* End of feature/parameter section, unless changing advanced things/modifying the sketch there's absolutely nothing to do further down! */


/* library, wifi and ntp stuff depending on above config/parameters */
#ifdef NODEMCU
  #if defined ( USENTP ) && !defined ( USEWIFI )                 // enforce USEWIFI when USENTP is defined
    #define USEWIFI
    #pragma warning "USENTP without USEWIFI, enabling WiFi"
  #endif
  #ifdef USEWIFI
    #include <ESP8266WiFi.h>
    #include <WiFiUdp.h>
  #endif
#endif

#ifdef USENTP
  #include <NTPClient.h>
  WiFiUDP ntpUDP;
  NTPClient timeClient(ntpUDP, NTPHOST, 0, 60000);
#endif
/* end library stuff */


/* setting feature combinations/options */
#if defined ( FASTFORWARD ) || defined ( CUSTOMHELPER )
  bool firstLoop = true;
  #ifdef USERTC
    #undef USERTC
  #endif
  #ifdef USEWIFI
    #undef USEWIFI
  #endif
  #ifdef USENTP
    #undef USENTP
  #endif
  #ifdef AUTODST
    #undef AUTODST
  #endif
#endif
/* setting feature combinations/options */


/* Start of FastLED/clock stuff */
#define LEDSTUFF
#ifdef LEDSTUFF
  #ifdef NODEMCU
    #define FASTLED_ESP8266_RAW_PIN_ORDER                        // this means we'll be using the raw esp8266 pin order -> GPIO_12, which is d6 on nodeMCU
    #define LED_PIN 12                                           // led data in connected to GPIO_12 (d6/nodeMCU)
  #else
    #define FASTLED_ALLOW_INTERRUPTS 0                           // AVR + WS2812 + IRQ = https://github.com/FastLED/FastLED/wiki/Interrupt-problems
    #define LED_PIN 6                                            // led data in connected to d6 (arduino)
  #endif
  
  #define LED_PWR_LIMIT 500                                      // 500mA - Power limit in mA (voltage is set in setup() to 5v)

  #ifdef LW_ENG
    #define LED_COUNT 100                                        // change if using additional leds to indicate am/pm!
  #endif

  #ifdef LW_GER
    #define LED_COUNT 106
  #endif
  
  #include <FastLED.h>
  
  uint8_t markerHSV[3] = { 96, 255, 100 };                       // this color will be used to "flag" leds for coloring later on while updating the leds
  CRGB leds[LED_COUNT];
  CRGBPalette16 currentPalette;
#endif


// start clock specific config/parameters

// ---- english ----

#ifdef LW_ENG
const uint8_t wordGroups[11][2] PROGMEM = {   // first and last LED used for words, lower one first
  {  95,  96 },   // "IT"           0
  {  98,  99 },   // "IS"           1
  {  78,  81 },   // "FIVE"         2
  {  68,  70 },   // "TEN"          3
  {  88,  94 },   // "QUARTER"      4
  {  72,  77 },   // "TWENTY"       5
  {  72,  81 },   // "TWENTYFIVE"   6
  {  84,  87 },   // "HALF"         7
  {  62,  63 },   // "TO"           8
  {  63,  66 },   // "PAST"         9
  {   0,   5 }    // "O´CLOCK"     10
};

const uint8_t hourGroups[13][2] PROGMEM = {   // first and last LED used for hours, lower one first
  { 255, 255 },   // dummy - so we can use index 1-12 for hours 1-12 instead of index 0-11
  {  45,  47 },   // "ONE"
  {  47,  49 },   // "TWO"
  {  51,  55 },   // "THREE"
  {  36,  39 },   // "FOUR"
  {   7,  10 },   // "FIVE"
  {  42,  44 },   // "SIX"
  {  27,  31 },   // "SEVEN"
  {  55,  59 },   // "EIGHT"
  {  24,  27 },   // "NINE"
  {  33,  35 },   // "TEN"
  {  16,  21 },   // "ELEVEN"
  {  11,  16 }    // "TWELVE"
};

const uint8_t displayContents[12][6] PROGMEM = {  // first value is current range, minutes() / 5 = 0-11
  {   0,   0,   1,  10, 255, 255 },    // minute  0 -  4, words 0, 1 and 10  - 255 equals blank/nothing
  {   1,   2,   9, 255, 255, 255 },    // minute  5 -  9, words 2 and 9
  {   2,   0,   1,   3,   9, 255 },    // minute 10 - 14
  {   3,   4,   9, 255, 255, 255 },    // minute 15 - 19
  {   4,   0,   1,   5,   9, 255 },    // minute 20 - 24
  {   5,   6,   9, 255, 255, 255 },    // minute 25 - 29
  {   6,   0,   1,   7,   9, 255 },    // minute 30 - 34
  {   7,   6,   8, 255, 255, 255 },    // minute 35 - 39
  {   8,   0,   1,   5,   8, 255 },    // minute 40 - 44
  {   9,   4,   8, 255, 255, 255 },    // minute 45 - 49
  {  10,   0,   1,   3,   8, 255 },    // minute 50 - 54
  {  11,   2,   8, 255, 255, 255 }     // minute 55 - 59
};
#endif

// ---- german ----

#ifdef LW_GER
const uint8_t wordGroups[10][2] PROGMEM = {   // first and last LED used for words, lower one first
  {  95,  96 },   // "ES"        0
  {  98, 100 },   // "IST"       1
  { 102, 105 },   // "FÜNF"      2
  {  91,  94 },   // "ZEHN"      3
  {  72,  78 },   // "VIERTEL"   4
  {  84,  90 },   // "ZWANZIG"   5
  {  51,  54 },   // "HALB"      6
  {  62,  64 },   // "VOR"       7
  {  67,  70 },   // "NACH"      8
  {   0,   2 }    // "UHR"       9
};

const uint8_t hourGroups[13][2] PROGMEM = {   // first and last LED used for hours, lower one first
  {  13,  15 },   // "EIN"      // only used at "ES IST EIN UHR", otherwise index 1-12 = hour
  {  13,  16 },   // "EINS"
  {  56,  59 },   // "ZWEI"
  {  11,  14 },   // "DREI"
  {  46,  49 },   // "VIER"
  {  24,  27 },   // "FÜNF"
  {  35,  39 },   // "SECHS"
  {  16,  21 },   // "SIEBEN"
  {  42,  45 },   // "ACHT"
  {   4,   7 },   // "NEUN"
  {   7,  10 },   // "ZEHN"
  {  32,  34 },   // "ELF"
  {  27,  31 }    // "ZWÖLF"
};

const uint8_t displayContents[12][6] PROGMEM = {  // first value is current range, minutes() / 5 = 0-11
  {   0,   0,   1,   9, 255, 255 },    // minute  0 -  4, words 0, 1 and 9  - 255 equals blank/nothing
  {   1,   2,   8, 255, 255, 255 },    // minute  5 -  9, words 2 and 8
  {   2,   0,   1,   3,   8, 255 },    // minute 10 - 14
  {   3,   4,   8, 255, 255, 255 },    // minute 15 - 19
  {   4,   0,   1,   5,   8, 255 },    // minute 20 - 24
  {   5,   2,   7,   6, 255, 255 },    // minute 25 - 29
  {   6,   0,   1,   6, 255, 255 },    // minute 30 - 34
  {   7,   2,   8,   6, 255, 255 },    // minute 35 - 39
  {   8,   0,   1,   5,   7, 255 },    // minute 40 - 44
  {   9,   4,   7, 255, 255, 255 },    // minute 45 - 49
  {  10,   0,   1,   3,   7, 255 },    // minute 50 - 54
  {  11,   2,   7, 255, 255, 255 }     // minute 55 - 59
};
#endif

uint8_t clockStatus = 1;                  // Used for various things, don't mess around with it! 1 = startup
                                          // 0 = regular mode, 1 = startup, 9x = setup modes (90, 91, 92, 93...), 80 = color preview mode on Lazy Words

/* these values will be saved to EEPROM:
  0 = index for selected palette
  1 = index for selected brightness level */

/* End of FastLED/clock stuff */
// End clock specific configs/parameters

/* other global variables */
uint8_t btnRepeatCounter = 0;         // keeps track of how often a button press has been repeated
/* */


/* -- this is where the fun parts start -------------------------------------------------------------------------------------------------------- */


void setup() {
  #ifdef DEBUG
    while ( millis() < 300 ) {  // safety delay for serial output
      #ifdef NODEMCU
        yield();
      #endif
    }
    #ifdef NODEMCU
      Serial.begin(74880); Serial.println(F("  "));
    #else
      Serial.begin(57600);  Serial.println(F("  "));
    #endif
    #ifdef SKETCHNAME
      Serial.print(SKETCHNAME); Serial.println(F(" starting up..."));
    #endif
    #ifdef CLOCKNAME
      Serial.print("Clock Type: "); Serial.println(CLOCKNAME);
    #endif
    #ifdef RTCTYPE
      Serial.print(F("Configured RTC: ")); Serial.println(RTCTYPE);
    #endif
    #ifdef LEDSTUFF
      Serial.print(F("LED power limit: ")); Serial.print(LED_PWR_LIMIT); Serial.println(F(" mA"));
      Serial.print(F("Total LED count: ")); Serial.println(LED_COUNT);
    #endif
    #ifdef AUTODST
      Serial.println(F("autoDST enabled"));
    #endif
    #ifdef LW_ENG
      Serial.println(F("Language: English"));
    #endif
    #ifdef LW_GER
      Serial.println(F("Language: German"));
    #endif
    #ifdef NODEMCU
      Serial.println(F("Configured for nodeMCU"));
      #ifdef USEWIFI
        Serial.println(F("WiFi enabled"));
      #endif
      #ifdef USENTP
        Serial.print(F("NTP enabled, NTPHOST: ")); Serial.println(NTPHOST);
      #endif
    #else
      Serial.println(F("Configured for Arduino"));
    #endif
    #ifdef FASTFORWARD
      Serial.println(F("!! FASTFORWARD defined !!"));
    #endif
    while ( millis() < 600 ) {  // safety delay for serial output
      #ifdef NODEMCU
        yield();
      #endif
    }
  #endif
  
  #ifdef AUTOBRIGHTNESS
    #ifdef DEBUG
      Serial.print(F("autoBrightness enabled, LDR using pin: ")); Serial.println(pinLDR);
    #endif
    pinMode(pinLDR, INPUT);
  #endif

  #ifdef LEDSTUFF
    FastLED.addLeds<WS2812B, LED_PIN, GRB>(leds, LED_COUNT).setCorrection(TypicalSMD5050).setTemperature(DirectSunlight).setDither(1);
    FastLED.setMaxPowerInVoltsAndMilliamps(5, LED_PWR_LIMIT);
    FastLED.clear();
    FastLED.show();
  #endif
  
  pinMode(buttonA, INPUT_PULLUP);
  pinMode(buttonB, INPUT_PULLUP);

  #ifdef DEBUG
    if ( digitalRead(buttonA) == LOW || digitalRead(buttonB) == LOW ) {
      if ( digitalRead(buttonA) == LOW ) {
        Serial.println(F("buttonA is LOW / pressed - check wiring!"));
      }
      if ( digitalRead(buttonB) == LOW ) {
        Serial.println(F("buttonB is LOW / pressed - check wiring!"));
        Serial.println(F("Lazy Words test display starting in 3 seconds..."));
        unsigned long testStart = millis();
        while ( millis() - testStart < 3000 ) {
          #ifdef NODEMCU
          yield();
          #endif
        }
        #ifdef LW_ENG
          showWord(0); showWord(1);
          showWord(6); showWord(8);
          showWord(10); showHour(12);
        #endif
        #ifdef LW_GER
          showWord(0); showWord(1);
          showWord(6); showWord(9);
          showHour(10); showHour(2);
        #endif
        for ( uint8_t i = 0; i < LED_COUNT; i++ ) {
          if ( leds[i] ) leds[i].setHSV(0, 160, 160);
        }
        FastLED.show();
        Serial.println(F("...done, reset/power cycle to restart."));
        while ( 0 < 1 ) {
          #ifdef NODEMCU
          yield();
          #endif
        }
      }
    }
  #endif

  #ifdef NODEMCU                                                                              // if building for nodeMCU...
    #ifdef USEWIFI                                                                            // ...and if using WiFi.....
      #ifdef DEBUG
        Serial.println(F("Starting up WiFi..."));
      #endif
      WiFi.mode(WIFI_STA);                                                                    // set WiFi mode to STA...
      if ( useWPS ) {
        WiFi.begin(WiFi.SSID().c_str(),WiFi.psk().c_str());                                   // ...and start connecting using saved credentials...
        #ifdef DEBUG
          Serial.println(F("Using WPS setup / saved credentials"));
        #endif
      } else {
        WiFi.begin(wifiSSID, wifiPWD);                                                        // ...or credentials defined in the USEWIFI config section
        #ifdef DEBUG
          Serial.println(F("Using credentials from sketch"));
        #endif
      }
      unsigned long startTimer = millis();
      uint8_t wlStatus = 0;
      uint8_t counter = 6;
      #ifdef DEBUG
        Serial.print(F("Waiting for WiFi connection... "));
      #endif
      while ( wlStatus == 0 ) {
        if ( WiFi.status() != WL_CONNECTED ) wlStatus = 0; else wlStatus = 1;
        #ifdef LEDSTUFF
          if ( millis() - startTimer >= 1000 ) {
            FastLED.clear();
            showHour(counter);
            FastLED.show();
            if ( counter > 0 ) counter--; else wlStatus = 2;
            startTimer = millis();
            #ifdef DEBUG
                Serial.print(F("."));
            #endif
          }
        #endif
        #ifdef NODEMCU
          yield();
        #endif
      }
      if ( WiFi.status() == WL_CONNECTED ) {                                                  // if status is connected...
        #ifdef USENTP                                                                         // ...and USENTP defined...
          timeClient.begin();                                                                 // ...start timeClient
        #endif
      }
      #ifdef DEBUG
        Serial.println();
        if ( WiFi.status() != 0 ) {
          Serial.print(F("setup(): Connected to SSID: ")); Serial.println(WiFi.SSID());
        } else Serial.println(F("setup(): WiFi connection failed."));
      #endif
    #endif
    EEPROM.begin(512);
  #endif

  #ifdef USERTC
    Rtc.Begin();
    if ( Rtc.GetIsRunning() == false ) {
      #ifdef DEBUG
        Serial.println(F("setup(): RTC not running, trying to start..."));
      #endif
      Rtc.SetIsRunning(true);
    }
    #ifdef DEBUG
      Serial.println(F("setup(): RTC.begin(), 2 second safety delay before"));
      Serial.println(F("         doing any read/write actions!"));
    #endif
    unsigned long tmp_time = millis();
    while ( millis() - tmp_time < 2000 ) {
      #ifdef NODEMCU
        yield();
      #endif
    }
    #ifdef DEBUG
      Serial.println(F("setup(): RTC initialized"));
    #endif
  #else
    #ifdef DEBUG
      Serial.println(F("setup(): No RTC defined!"));
      setTime(12, 0, 0, 1, 1, 2000);
    #endif
  #endif

  #ifdef LEDSTUFF
    FastLED.clear();
    FastLED.show();
    /* eeprom settings */
    #ifdef nodeMCU
      EEPROM.begin(512);
    #endif
    paletteSwitcher();
    brightnessSwitcher();
  #endif

  #ifdef FASTFORWARD
    setTime(21, 59, 50, 31, 1, 2023);            // h, m, s, d, m, y to set the clock to when using FASTFORWARD
  #endif

  #ifdef USENTP
    syncHelper(); 
  #endif

  clockStatus = 0;       // change from 1 (startup) to 0 (running mode)

  #ifdef DEBUG
    printTime();
    Serial.println(F("setup() done"));
    Serial.println(F("------------------------------------------------------"));
  #endif
}


/* MAIN LOOP */


void loop() {
  static uint8_t lastInput = 0;                     // != 0 if any button press has been detected
  static uint8_t lastSecondDisplayed = 0;           // This keeps track of the last second when the display was updated (HH:MM and HH:MM:SS)
  static unsigned long lastCheckRTC = millis();     // This will be used to read system time in case no RTC is defined (not supported!)
  static bool doUpdate = false;                     // Update led content whenever something sets this to true. Coloring will always happen at fixed intervals!
  #ifdef USERTC
    static RtcDateTime rtcTime = Rtc.GetDateTime(); // Get time from rtc
  #else
    static time_t sysTime = now();                  // if no rtc is defined, get local system time
  #endif
  #ifdef LEDSTUFF
    static uint8_t refreshDelay = 5;                // refresh leds every 5ms
    static long lastRefresh = millis();             // Keeps track of the last led update/FastLED.show() inside the loop
    #ifdef AUTOBRIGHTNESS
    static long lastReadLDR = millis();
    #endif
  #endif
  #ifdef FASTFORWARD
    static unsigned long lastFFStep = millis();     // Keeps track of last time increment if FASTFORWARD is defined
  #endif
  
  if ( lastInput != 0 ) {                                                                  // If any button press is detected...
    if ( btnRepeatCounter < 1 ) {                                                          // execute short/single press function(s)
      #ifdef DEBUG
        Serial.print(F("loop(): ")); Serial.print(lastInput); Serial.println(F(" (short press)"));
      #endif
      if ( lastInput == 1 ) {                                                              // short press button A
        #ifdef LEDSTUFF
          brightnessSwitcher();
        #endif
      }
      if ( lastInput == 2 ) {                                                              // short press button B
        #ifdef LEDSTUFF                                                                    // improvised preview on Lazy Words
          paletteSwitcher();
          if ( colorPreview ) {
            unsigned long paletteChanged = millis();
            uint16_t colorSpeedBak = colorSpeed;
            colorSpeed = 3;
            #ifdef DEBUG
              Serial.print(F("Palette changed, previewing..."));
            #endif
            clockStatus = 80;
            while ( millis() - paletteChanged <= colorPreviewDuration * 1000 ) {
              #ifdef USERTC
                displayTime(now());
              #else
                displayTime(sysTime);
              #endif
              colorizeOutput(colorMode);
              FastLED.show();
              FastLED.clear();
              #ifdef NODEMCU
                yield();
              #endif
            }
            #ifdef DEBUG
              Serial.println(F("...done!"));
            #endif
            colorSpeed = colorSpeedBak;
            clockStatus = 0;            
          }
        #endif
      }
      if ( lastInput == 3 ) {                                                              // short press button A + button B
      }
    } else if ( btnRepeatCounter > 8 ) {                                                   // execute long press function(s)...
      btnRepeatCounter = 1;                                                                // ..reset btnRepeatCounter to stop this from repeating
      #ifdef DEBUG
        Serial.print(F("loop(): ")); Serial.print(lastInput); Serial.println(F(" (long press)"));
      #endif
      if ( lastInput == 1 ) {                                                              // long press button A, set hour - 1
        Serial.println(F("Waiting for RTC..."));
        while ( second(Rtc.GetDateTime().Epoch32Time() ) == lastSecondDisplayed ) {
          // waiting for RTC
          #ifdef NODEMCU
            yield();
          #endif
        }
        Rtc.SetDateTime( (Rtc.GetDateTime() - 3600) );
        #ifdef DEBUG
          Serial.println(F("Time changed by -1 hour"));
        #endif
        FastLED.clear();
        displayTime(Rtc.GetDateTime().Epoch32Time());
        colorizeOutput(colorMode);
        FastLED.show();
        #ifdef FADING
          pixelFader();
        #endif
      }
      if ( lastInput == 2 ) {                                                              // long press button B, set hour + 1
        Serial.println(F("Waiting for RTC..."));
        while ( second(Rtc.GetDateTime().Epoch32Time() ) == lastSecondDisplayed ) {
          // waiting for RTC   
          #ifdef NODEMCU
            yield();
          #endif
        }
        Rtc.SetDateTime( (Rtc.GetDateTime() + 3600) );
        #ifdef DEBUG
          Serial.println(F("Time changed by +1 hour"));
        #endif
        FastLED.clear();
        displayTime(Rtc.GetDateTime().Epoch32Time());
        colorizeOutput(colorMode);
        FastLED.show();
        #ifdef FADING
          pixelFader();
        #endif
      }
      if ( lastInput == 3) {                                                               // long press button A + button B
        #ifdef USEWIFI                                                                     // if USEWIFI is defined and...
          // nothing to do here, wps not available on Lazy Words
        #else                                                                              // if USEWIFI is not defined...
          #ifdef LEDSTUFF
            FastLED.clear();
            FastLED.show();
            setupClock();                                                                  // start time setup
          #endif
        #endif
      }
      while ( digitalRead(buttonA) == LOW || digitalRead(buttonB) == LOW ) {               // wait until buttons are released again
        #ifdef LEDSTUFF
          if ( millis() % 50 == 0 ) {                                                      // Refresh leds every 50ms to give optical feedback
            colorizeOutput(colorMode);
            FastLED.show();
          }
        #endif
        #ifdef NODEMCU
          yield();
        #endif
      }
    }
  }
  
  #ifdef FASTFORWARD                                                                       // if FASTFORWARD is defined...
    if ( millis() - lastFFStep >= 100 ) {                                                  // ...and 100ms have passed...
      adjustTime(15);                                                                      // ...add 15 seconds to current time
      lastFFStep = millis();
    }
  #endif

  if ( millis() - lastCheckRTC >= 50 ) {                                                   // check rtc/system time every 50ms
    #ifdef USERTC
      rtcTime = Rtc.GetDateTime();
      if ( lastSecondDisplayed != rtcTime.Second() ) doUpdate = true;
    #else
      sysTime = now();
      if ( lastSecondDisplayed != second(sysTime) ) doUpdate = true;
    #endif
    lastCheckRTC = millis();
  }

  if ( doUpdate ) {                                      // this will update the led array if doUpdate is true because of a new second from the rtc
    #ifdef USERTC
      setTime(rtcTime.Hour(), rtcTime.Minute(), rtcTime.Second(),
              rtcTime.Day(), rtcTime.Month(), rtcTime.Year() );      // sync system time to rtc every second
      #ifdef LEDSTUFF
        FastLED.clear();                                 // 1A - clear all leds...
        displayTime(now());                              // 2A - output rtcTime to the led array..
      #endif
      lastSecondDisplayed = rtcTime.Second();
    #else
      #ifdef LEDSTUFF
        FastLED.clear();                                 // 1B - clear all leds...
        displayTime(now());                              // 2B - output sysTime to the led array...
      #endif
      lastSecondDisplayed = second(sysTime); 
    #endif
    doUpdate = false;
    #ifdef DEBUG
      if ( second() % 20 == 0 ) {
        printTime();
      }
    #endif
    #ifdef USENTP                                                                          // if NTP is enabled, resync to ntp server at 3:01:00 am
      time_t t = myTimeZone.toLocal(now());                                                // convert current system time to local time zone according to rules on top of the sketch
      if ( hour(t) == 3 && minute(t) == 1 and second(t) == 0 ) {
        syncHelper();
      }
    #endif
  }

  #ifdef LEDSTUFF
    colorizeOutput(colorMode);                           // 1C, 2C, 3C...colorize the data inside the led array right now...
    #ifdef AUTOBRIGHTNESS
      if ( millis() - lastReadLDR >= intervalLDR ) {     // if LDR is enabled and sample interval has been reached...
        readLDR();                                       // ...call readLDR();
        if ( abs(avgLDR - lastAvgLDR) >= 5 ) {           // if avgLDR has changed for more than +/- 5 update lastAvgLDR
          lastAvgLDR = avgLDR;
          FastLED.setBrightness(avgLDR);
        }
        lastReadLDR = millis();
      }
    #endif
    #ifdef FADING
      pixelFader();
    #endif
    if ( millis() - lastRefresh >= refreshDelay ) {
      FastLED.show();
      lastRefresh = millis();
    }
  #endif

  lastInput = inputButtons();
}


/* */


#ifdef LEDSTUFF

void showWord(uint8_t w) {
  if ( w >= sizeof(wordGroups) ) return;
  uint8_t startLED = pgm_read_byte_near(&wordGroups[w][0]);
  uint8_t endLED = pgm_read_byte_near(&wordGroups[w][1]);
  for ( uint8_t i = startLED; i <= endLED; i++ ) {
    if ( i == 255 ) return; else
      leds[i].setHSV(markerHSV[0], markerHSV[1], markerHSV[2]);
  }
}

void showHour(uint8_t h) {
  if ( h >= sizeof(hourGroups) ) return;
  uint8_t startLED = pgm_read_byte_near(&hourGroups[h][0]);
  uint8_t endLED = pgm_read_byte_near(&hourGroups[h][1]);
  for ( uint8_t i = startLED; i <= endLED; i++ ) {
    if ( i == 255 ) return; else
      leds[i].setHSV(markerHSV[0], markerHSV[1], markerHSV[2]);
  }
}

#ifdef FADING
void fadePixel(uint8_t i, uint8_t amount, uint8_t fadeType) {
  /* this will check if the first led of a given segment is lit and if it is, will fade by  
     amount using fadeType. fadeType is important because when fading things in that where
     off previously we must avoid setting them black at first - hence fadeLightBy instead
     of fadeToBlack.  */
  if ( leds[i] ) {
    if ( fadeType == 0 ) {
      leds[i].fadeToBlackBy(amount);
    } else {
      leds[i].fadeLightBy(amount);
    }
  }
}

void pixelFader() {
  if ( fadePixels == 0 ) return;
  static unsigned long firstRun = 0;                                                                // time when a change has been detected and fading starts
  static unsigned long lastRun = 0;                                                                 // used to store time when this function was executed the last time
  static boolean active = false;                                                                    // will be used as a flag when to do something / fade pixels
  static uint8_t previousPixels[LED_COUNT] = { 0 };                                                 // all the pixels lit after the last run
  static uint8_t currentPixels[LED_COUNT] = { 0 };                                                  // all the pixels lit right now
  static uint8_t changedPixels[LED_COUNT] = { 0 };                                                  // used to store the differences -> 1 = led has been turned off, fade out, 2 = was off, fade in
  static uint8_t fadeSteps = 15;                                                                    // steps used to fade in or out
  lastRun = millis();
  if ( !active ) {                                                                                  // this will check if....
    firstRun = millis();
    for ( uint8_t i = 0; i < LED_COUNT; i++ ) {                                                     // ...any of the pixels are on....
      if ( leds[i] ) {
        currentPixels[i] = 1;
      } else {
        currentPixels[i] = 0;
      }
      if ( currentPixels[i] != previousPixels[i] ) {                                                // ...and compare them to the previous displayed pixels.
        active = true;                                                                              // if a change has been detected, set active = true so fading gets executed
        if ( currentPixels[i] == 0 ) {
          changedPixels[i] = 1;
        } else {
          changedPixels[i] = 2;
        }
      }
    }
  }
  if ( active ) {                                                                                   // this part is executed once a change has been detected....
    static uint8_t counter = 1;
    static unsigned long lastFadeStep = millis();
    for ( uint8_t i = 0; i < LED_COUNT; i++ ) {                                                     // redraw pixels that have turned off, so we can fade them out...
      if ( changedPixels[i] == 1 ) {
          leds[i].setHSV(markerHSV[0], markerHSV[1], markerHSV[2]);
      }
    }
    colorizeOutput(colorMode);                                                                      // colorize again after redraw, so colors keep consistent
    for ( uint8_t i = 0; i < LED_COUNT; i++ ) {
      if ( changedPixels[i] == 1 ) {                                                                // 1 - pixel has turned on, this one has to be faded in
        fadePixel(i, counter * ( 255.0 / fadeSteps ), 0);                                           // fadeToBlackBy, pixels supposed to be off/fading out
      }
      if ( changedPixels[i] == 2 ) {                                                                // 2 - pixel has turned off, this one has to be faded out
        if ( fadePixels == 2 ) {
          fadePixel(i,255 - counter * ( 255.0 / fadeSteps ), 1 );                                   // fadeLightBy, pixels supposed to be on/fading in
        }
      }
    }
    if ( millis() - lastFadeStep >= fadeDelay ) {
      counter++;
      lastFadeStep = millis();
    }
    if ( counter > fadeSteps ) {                                                                    // done with fading, reset variables...
      counter = 1;
      active = false;
      for ( uint8_t pixel = 0; pixel < LED_COUNT; pixel++ ) {                                       // and save current pixels to previousPixels
        if ( leds[pixel] ) {
          previousPixels[pixel] = 1;
        } else {
          previousPixels[pixel] = 0;
        }
        changedPixels[pixel] = 0;
      }
      #ifdef DEBUG
        Serial.print(F("pixel fading sequence took "));                                             // for debugging/checking duration - fading should never take longer than 1000ms!
        Serial.print(millis() - firstRun);
        Serial.println(F(" ms"));
      #endif
    }
  }
}
#endif


#ifdef AUTOBRIGHTNESS
void readLDR() {                                                                                            // read LDR value 5 times and write average to avgLDR
  static uint8_t runCounter = 1;
  static uint16_t tmp = 0;
  uint8_t readOut = map(analogRead(pinLDR), 0, 1023, 0, 250);
  tmp += readOut;
  if (runCounter == 5) {
    avgLDR = ( tmp / 5 )  * factorLDR;
    tmp = 0; runCounter = 0;
    #ifdef DEBUG
      if ( dbgLDR ) {
        Serial.print(F("readLDR(): avgLDR value: "));
        Serial.print(avgLDR);
      }
    #endif
    if ( avgLDR < minBrightness ) avgLDR = minBrightness;
    if ( avgLDR > brightness ) avgLDR = brightness;
    if ( avgLDR >= upperLimitLDR && avgLDR < brightness ) avgLDR = brightness;                              // if avgLDR is above upperLimitLDR switch to max current brightness
    if ( avgLDR <= lowerLimitLDR ) avgLDR = minBrightness;                                                  // if avgLDR is below lowerLimitLDR switch to minBrightness
    #ifdef DEBUG
      if ( dbgLDR ) {
        Serial.print(F(" - adjusted to: "));
        Serial.println(avgLDR);
      }
    #endif
  }
  runCounter++;
}
#endif

void setupClock() {
/* This sets time and date (if AUTODST is defined) on the clock/rtc */
  clockStatus = 90;                                                                         // clockStatus 9x = setup, relevant for other functions/coloring
  while ( digitalRead(buttonA) == LOW || digitalRead(buttonB) == LOW ) {                    // do nothing until both buttons are released to avoid accidental inputs right away
    #ifdef NODEMCU
      yield();
    #endif
  }
  tmElements_t setupTime;                                                                   // Create a time element which will be used. Using the current time would
  setupTime.Hour = 12;                                                                      // give some problems (like time still running while setting hours/minutes)
  setupTime.Minute = 0;                                                                     // Setup starts at 12 (12 pm) (utc 12 if AUTODST is defined)
  setupTime.Second = 0;                                                                     // 
  setupTime.Day = 1;                                                                        // date settings only used when AUTODST is defined, but will set them anyways
  setupTime.Month = 2;                                                                      // see above
  setupTime.Year = 23;                                                                      // current year - 1970 (2022 - 1970 = 52) -- date cannot be set on Lazy Words manually!
  #ifdef USERTC
    RtcDateTime writeTime;
  #endif
  uint8_t lastInput = 0;
  // minutes
  while ( lastInput != 2 ) {
    clockStatus = 93;
    if ( lastInput == 1 ) {
      if ( setupTime.Minute < 55 ) {
        setupTime.Minute+=5;
      } else {
        setupTime.Minute = 0;
        if ( setupTime.Hour < 23 ) {
          setupTime.Hour++;
        } else {
          setupTime.Hour = 0;
        }
      }
      #ifdef DEBUG
        Serial.print(F("setupClock(): "));
        Serial.print(hourFormat12(makeTime(setupTime))); Serial.print(F("h")); 
        Serial.print(setupTime.Minute); Serial.println(F("m"));
      #endif
    }
    displayTime(makeTime(setupTime));
    colorizeOutput(colorMode);
    FastLED.show();
    lastInput = inputButtons();
  }
  lastInput = 0;
  #ifdef DEBUG
    Serial.print(F("setupClock(): "));
    Serial.print(F("HH:MM:SS -> "));
    if ( setupTime.Hour < 10 ) Serial.print(F("0"));
    Serial.print(setupTime.Hour); Serial.print(F(":"));
    if ( setupTime.Minute < 10 ) Serial.print(F("0"));
    Serial.print(setupTime.Minute); Serial.print(F(":"));
    if ( setupTime.Second < 10 ) Serial.print(F("0"));
    Serial.println(setupTime.Second);
  #endif
  #ifdef USERTC
    writeTime = { 2000 + setupTime.Year, setupTime.Month, setupTime.Day,
                  setupTime.Hour, setupTime.Minute, setupTime.Second };
    Rtc.SetDateTime(writeTime);
    setTime(writeTime.Hour(), writeTime.Minute(), writeTime.Second(), writeTime.Day(), writeTime.Month(), writeTime.Year());
    #ifdef DEBUG
      Serial.println(F("setupClock(): RTC time set"));
      Serial.println(makeTime(setupTime));
      printTime();
    #endif
  #else
      setupTime.Year += 30;
      setTime(makeTime(setupTime));
  #endif
  clockStatus = 0;
  #ifdef DEBUG
    Serial.println(F("setupClock() done"));
  #endif
}


void colorizeOutput(uint8_t mode) {
/* So far showWord()/showHour() only set some leds inside the array to values from "markerHSV" but we haven't updated
   the leds yet using FastLED.show(). This function does the coloring of the right now single colored but "insivible"
   output. This way color updates/cycles aren't tied to updating display contents                                      */
  static unsigned long lastColorChange = 0;
  static uint8_t startColor = 0;
  static uint8_t colorOffset = 4;              // different offsets result in quite different results, depending on the amount of leds inside each segment...
                                               // ...so it's set inside each color mode if required
  if ( clockStatus == 80 ) {                   // 80 -> color preview cylcing after switching in loop()
    colorOffset = 20;  
  } else {
    colorOffset = 4;
  }
  
  // mode 0 = check every segment if it's lit and assign a color based on led #
  if ( mode == 0 ) {
    for ( uint8_t i = 0; i < LED_COUNT; i++ ) {
      if ( leds[i] ) leds[i] = ColorFromPalette(currentPalette, startColor + i * colorOffset, brightness, LINEARBLEND);
    }
  }

  /* clockStatus >= 90 is used for coloring output while in setup mode */
  if ( clockStatus >= 90 ) {
    static boolean blinkFlag = true;
    static unsigned long lastBlink = millis();
    static uint8_t b = brightnessLevels[0];
    if ( millis() - lastBlink > 333 ) {                    // blink switch frequency, 3 times a second
      if ( blinkFlag ) {
        blinkFlag = false;
        b = brightnessLevels[1];
      } else {
        blinkFlag = true;
        b = brightnessLevels[0];
      }
      lastBlink = millis();
    }                                                      // unset values = red, set value = green, current value = yellow and blinkinkg 
    for ( uint8_t i = 0; i < LED_COUNT; i++ ) {
      if ( leds[i] ) leds[i].setHSV(0, 255, b);
    }
  }

  #ifdef FASTFORWARD
    if ( millis() - lastColorChange > 30 ) {
  #else
    if ( millis() - lastColorChange > colorSpeed ) {
  #endif
    if ( reverseColorCycling ) {
      startColor--;
    } else {
      startColor++;
    }
    lastColorChange = millis();
  }
  
  #ifdef AUTOBRIGHTNESS
  if ( nightMode && clockStatus == 0 ) {                           // nightmode will overwrite everything that has happened so far...
    for ( uint16_t i = 0; i < LED_COUNT; i++ ) {
      if ( leds[i] ) {
        if ( avgLDR == minBrightness ) {
          leds[i].setHSV(nightColor[0], 255, nightColor[1] );      // and assign nightColor to all lit leds. Default is a very dark red.
          FastLED.setDither(0);
        } else {
          FastLED.setDither(1);
        }
      }
    }
  }
  #endif
}


void displayTime(time_t t) {
  if ( clockStatus >= 90 ) {  // while in setup this will clear the display each time when redrawing
    FastLED.clear();
  }
  
  for ( uint8_t i = 1; i < 6; i++ ) {
    if ( pgm_read_byte_near(&displayContents[minute(t)/5][i]) != 255 ) {
      showWord(pgm_read_byte_near(&displayContents[minute(t)/5][i]));
    }
  }

  #ifdef LW_ENG
  if ( minute(t) / 5 < 7 ) {
    showHour(hourFormat12(t));
  } else {
    showHour(hourFormat12(t + 3600));
  }
  #endif

  #ifdef LW_GER
  if ( minute(t) / 5 < 5 ) {
    if ( hourFormat12(t) == 1 && minute(t) / 5 == 0 ) {  // Wenn "Ein Uhr", nutze "Ein" statt "Eins" aus hourGroups
      showHour(0);
    } else {
      showHour(hourFormat12(t));
    }
  } else {
    showHour(hourFormat12(t + 3600));
  }
  #endif
}


void paletteSwitcher() {    
/* As the name suggests this takes care of switching palettes. When adding palettes, make sure paletteCount increases
  accordingly.  A few examples of gradients/solid colors by using RGB values or HTML Color Codes  below               */
  static uint8_t paletteCount = 6;                                          
  static uint8_t currentIndex = 0;
  if ( clockStatus == 1 ) {                                                 // Clock is starting up, so load selected palette from eeprom...
    uint8_t tmp = EEPROM.read(0);
    if ( tmp >= 0 && tmp < paletteCount ) {
      currentIndex = tmp;                                                   // 255 from eeprom would mean there's nothing been written yet, so checking range...
    } else {
      currentIndex = 0;                                                     // ...and default to 0 if returned value from eeprom is not 0 - 6
    }
    #ifdef DEBUG
      Serial.print(F("paletteSwitcher(): loaded EEPROM value "));
      Serial.println(tmp);
    #endif
  }
  switch ( currentIndex ) {
    case 0: currentPalette = CRGBPalette16( CRGB( 224,   0,  40 ),
                                            CRGB(   8,   0, 244 ),
                                            CRGB( 100,   0, 180 ),
                                            CRGB( 208,   0,  96 ) ); break;
    case 1: currentPalette = CRGBPalette16( CRGB( 224,  16,   0 ),
                                            CRGB( 192,  64,   0 ),
                                            CRGB( 192, 140,   0 ),
                                            CRGB( 240,  40,   0 ) ); break;
    case 2: currentPalette = CRGBPalette16( CRGB::Aquamarine,
                                            CRGB::Turquoise,
                                            CRGB::Blue,
                                            CRGB::DeepSkyBlue   ); break;
    case 3: currentPalette = RainbowColors_p; break;
    case 4: currentPalette = PartyColors_p; break;
    case 5: currentPalette = CRGBPalette16( CRGB::LawnGreen ); break;
  }
  #ifdef DEBUG
    Serial.print(F("paletteSwitcher(): selected palette "));
    Serial.println(currentIndex);
  #endif
  if ( clockStatus == 0 ) {                             // only save selected palette to eeprom if clock is in normal running mode, not while in startup/setup/whatever
    EEPROM.put(0, currentIndex);
    #ifdef NODEMCU
      EEPROM.commit();
    #endif
    #ifdef DEBUG
      Serial.print(F("paletteSwitcher(): saved index "));
      Serial.print(currentIndex);
      Serial.println(F(" to eeprom"));
    #endif
  }
  if ( currentIndex < paletteCount - 1 ) {
    currentIndex++;    
  } else {
    currentIndex = 0;
  }
  #ifdef DEBUG
    Serial.println(F("paletteSwitcher() done"));
  #endif
}


void brightnessSwitcher() {
  static uint8_t currentIndex = 0;
  if ( clockStatus == 1 ) {                                                 // Clock is starting up, so load selected palette from eeprom...
    uint8_t tmp = EEPROM.read(1);
    if ( tmp >= 0 && tmp < 3 ) {
      currentIndex = tmp;                                                   // 255 from eeprom would mean there's nothing been written yet, so checking range...
    } else {
      currentIndex = 0;                                                     // ...and default to 0 if returned value from eeprom is not 0 - 2
    }
    #ifdef DEBUG
      Serial.print(F("brightnessSwitcher(): loaded EEPROM value "));
      Serial.println(tmp);
    #endif
  }
  switch ( currentIndex ) {
    case 0: brightness = brightnessLevels[currentIndex]; break;
    case 1: brightness = brightnessLevels[currentIndex]; break;
    case 2: brightness = brightnessLevels[currentIndex]; break;
  }
  #ifdef DEBUG
    Serial.print(F("brightnessSwitcher(): selected brightness index "));
    Serial.println(currentIndex);
  #endif
  if ( clockStatus == 0 ) {                             // only save selected brightness to eeprom if clock is in normal running mode, not while in startup/setup/whatever
    EEPROM.put(1, currentIndex);
    #ifdef NODEMCU
      EEPROM.commit();
    #endif
    #ifdef DEBUG
      Serial.print(F("brightnessSwitcher(): saved index "));
      Serial.print(currentIndex);
      Serial.println(F(" to eeprom"));
    #endif
  }
  if ( currentIndex < 2 ) {
    currentIndex++;    
  } else {
    currentIndex = 0;
  }
  #ifdef DEBUG
    Serial.println(F("brightnessSwitcher() done"));
  #endif
}


#endif /* LEDSTUFF */


bool leapYear(uint16_t y) {
  boolean isLeapYear = false;
  if (y % 4 == 0) isLeapYear = true;
  if (y % 100 == 0 && y % 400 != 0) isLeapYear = false;
  if (y % 400 == 0) isLeapYear = true;
  if ( isLeapYear ) return true; else return false;
}


uint8_t inputButtons() {
  /* This scans for button presses and keeps track of delay/repeat for user inputs
     Short keypresses will only be returned when buttons are released before repeatDelay
     is reached. This is to avoid constantly sending 1 or 2 when executing a long button
     press and/or multiple buttons.
     Note: Buttons are using pinMode INPUT_PULLUP, so HIGH = not pressed, LOW = pressed! */
  static uint8_t scanInterval = 30;                                 // only check buttons every 30ms
  static uint16_t repeatDelay = 1000;                               // delay in milliseconds before repeating detected keypresses
  static uint8_t repeatRate = 1000 / 7;                             // 7 chars per 1000 milliseconds
  static uint8_t minTime = scanInterval * 2;                        // minimum time to register a button as pressed
  static unsigned long lastReadout = millis();                      // keeps track of when the last readout happened
  static unsigned long lastReturn = millis();                       // keeps track of when the last readout value was returned
  static uint8_t lastState = 0;                                     // button state from previous scan
  uint8_t currentState = 0;                                         // button state from current scan
  uint8_t retVal = 0;                                               // return value, will be 0 if no button is pressed
  static unsigned long eventStart = millis();                       // keep track of when button states are changing
  if ( millis() - lastReadout < scanInterval ) return 0;            // only scan for button presses every <scanInterval> ms
  if ( digitalRead(buttonA) == LOW ) currentState += 1;
  if ( digitalRead(buttonB) == LOW ) currentState += 2;
  if ( currentState == 0 && currentState == lastState ) {
    btnRepeatCounter = 0;
  }
  if ( currentState != 0 && currentState != lastState ) {           // if any button is pressed and different from the previous scan...
    eventStart = millis();                                          // ...reset eventStart to current time
    btnRepeatCounter = 0;                                           // ...and reset global variable btnRepeatCounter
  }
  if ( currentState != 0 && currentState == lastState ) {           // if same input has been detected at least twice (2x scanInterval)...
    if ( millis() - eventStart >= repeatDelay ) {                   // ...and longer than repeatDelay...
      if ( millis() - lastReturn >= repeatRate ) {                  // ...check for repeatRate...
        retVal = currentState;                                      // ...and set retVal to currentState
        btnRepeatCounter++;
        lastReturn = millis();
      } else retVal = 0;                                            // return 0 if repeatDelay hasn't been reached yet
    }
  }
  if ( currentState == 0 && currentState != lastState && millis() - eventStart >= minTime && btnRepeatCounter == 0 ) {
    retVal = lastState;                                             // return lastState if all buttons are released after having been pressed for <minTime> ms
    btnRepeatCounter = 0;
  }
  lastState = currentState;
  lastReadout = millis();
  #ifdef DEBUG                                                      // output some information and read serial input, if available
    uint8_t serialInput = dbgInput();
    if ( serialInput != 0 ) {
      //Serial.print(F("inputButtons(): Serial input detected: ")); Serial.println(serialInput);
      retVal = serialInput;
    }
    if ( retVal != 0 ){
      Serial.print(F("inputButtons(): Return value is: ")) ; Serial.print(retVal); Serial.print(F(" - btnRepeatCounter is: ")); Serial.println(btnRepeatCounter);
    }
  #endif
  return retVal;
}


// following will only be included if USENTP is defined
#ifdef USENTP
/* This syncs system time to the RTC at startup and will periodically do other sync related
   things, like syncing rtc to ntp time */
  void syncHelper() {
    static unsigned long lastSync = millis();    // keeps track of the last time a sync attempt has been made
    if ( millis() - lastSync < 60000 && clockStatus != 1 ) return;   // only allow one ntp request per minute
    if ( WiFi.status() != WL_CONNECTED ) {
      #ifdef DEBUG
        Serial.println(F("syncHelper(): No WiFi connection"));
        return;
      #endif
    }
    #ifndef USERTC
      #ifndef USENTP
        #ifdef DEBUG
          Serial.println(F("syncHelper(): No RTC and no NTP configured, nothing to do..."));
          return;
        #endif
      #endif
    #endif
    time_t ntpTime = 0;
    #ifdef USERTC
      RtcDateTime ntpTimeConverted;
    #endif
    if ( clockStatus == 1 ) {                                                                    // looks like the sketch has just started running...
      #ifdef DEBUG
        Serial.println(F("syncHelper(): Initial sync on power up..."));
      #endif
      ntpTime = getTimeNTP();
      #ifdef DEBUG
        Serial.print(F("syncHelper(): NTP result is "));
        Serial.println(ntpTime);
      #endif
      lastSync = millis();
    } else {
      #ifdef DEBUG
        Serial.println(F("syncHelper(): Resyncing to NTP..."));
      #endif
      ntpTime = getTimeNTP();
      #ifdef DEBUG
        Serial.print(F("syncHelper(): NTP result is "));
        Serial.println(ntpTime);
      #endif
      lastSync = millis();
    }
    #ifdef USERTC
      ntpTimeConverted = { year(ntpTime), month(ntpTime), day(ntpTime),
                           hour(ntpTime), minute(ntpTime), second(ntpTime) };
      RtcDateTime rtcTime = Rtc.GetDateTime();                                                 // get current time from the rtc....
      if ( ntpTime > 100 ) {
        Rtc.SetDateTime(ntpTimeConverted);
      }
    #else
      time_t sysTime = now();                                                                  // ...or from system
      #ifdef DEBUG
        Serial.println(F("syncHelper(): No RTC configured, using system time"));
        Serial.print(F("syncHelper(): sysTime was "));
        Serial.println(now());
      #endif
      if ( ntpTime > 100 ) { 
        setTime(ntpTime);
      }
    #endif
    #ifdef DEBUG
      Serial.println(F("syncHelper() done"));
    #endif
  }

  
  time_t getTimeNTP() {
    unsigned long startTime = millis();
    time_t timeNTP;
    if ( WiFi.status() != WL_CONNECTED ) {
      #ifdef DEBUG
        Serial.print(F("getTimeNTP(): Not connected, WiFi.status is "));
        Serial.println(WiFi.status());
      #endif
    }                                                                                         // Sometimes the connection doesn't work right away although status is WL_CONNECTED...
    while ( millis() - startTime < 2000 ) {                                                   // ...so we'll wait a moment before causing network traffic
      #ifdef NODEMCU
        yield();
      #endif
    }
    if ( timeClient.update() == true ) {
      timeNTP = timeClient.getEpochTime();
    } else {
      timeNTP = 1;
      #ifdef DEBUG
        Serial.print(F("getTimeNTP(): No valid ntp response: ")); Serial.println(timeNTP);
      #endif
    }
    #ifdef DEBUG
      Serial.println(F("getTimeNTP() done"));
    #endif
    return timeNTP;
  }
#endif
// ---


// functions below will only be included if DEBUG is defined on top of the sketch
#ifdef DEBUG
  void printTime() {
    /* outputs current system and RTC time to the serial monitor, adds autoDST if defined */
    time_t tmp = now();
    #ifdef USERTC
      RtcDateTime tmp2 = Rtc.GetDateTime();
      setTime(tmp2.Hour(), tmp2.Minute(), tmp2.Second(),
              tmp2.Day(), tmp2.Month(), tmp2.Year() );
      tmp = now();
    #endif
    Serial.println(F("-----------------------------------"));
    Serial.print(F("System time is : "));
    if ( hour(tmp) < 10 ) Serial.print(F("0"));
    Serial.print(hour(tmp)); Serial.print(F(":"));
    if ( minute(tmp) < 10 ) Serial.print(F("0"));
    Serial.print(minute(tmp)); Serial.print(F(":"));
    if ( second(tmp) < 10 ) Serial.print(F("0"));
    Serial.println(second(tmp));
    Serial.print(F("System date is : "));
    Serial.print(year(tmp)); Serial.print("-");
    Serial.print(month(tmp)); Serial.print("-");
    Serial.print(day(tmp)); Serial.println(F(" (Y/M/D)"));
    #ifdef USERTC
      Serial.print(F("RTC time is    : "));
      if ( tmp2.Hour() < 10 ) Serial.print(F("0"));
      Serial.print(tmp2.Hour()); Serial.print(F(":"));
      if ( tmp2.Minute() < 10 ) Serial.print(F("0"));
      Serial.print(tmp2.Minute()); Serial.print(F(":"));
      if ( tmp2.Second() < 10 ) Serial.print(F("0"));
      Serial.println(tmp2.Second());
      Serial.print(F("RTC date is    : "));
      Serial.print( tmp2.Year()); Serial.print("-");
      Serial.print( tmp2.Month()); Serial.print("-");
      Serial.print( tmp2.Day()); Serial.println(F(" (Y/M/D)"));
    #endif
    #ifdef AUTODST
      tmp = myTimeZone.toLocal(tmp);
      Serial.print(F("autoDST time is: "));
      if ( hour(tmp) < 10 ) Serial.print(F("0"));
      Serial.print(hour(tmp)); Serial.print(F(":"));
      if ( minute(tmp) < 10 ) Serial.print(F("0"));
      Serial.print(minute(tmp)); Serial.print(F(":"));
      if ( second(tmp) < 10 ) Serial.print(F("0"));
      Serial.println(second(tmp));
      Serial.print(F("autoDST date is: "));
      Serial.print(year(tmp)); Serial.print("-");
      Serial.print(month(tmp)); Serial.print("-");
      Serial.print(day(tmp)); Serial.println(F(" (Y/M/D)"));
    #endif
    Serial.println(F("-----------------------------------"));
  }

  
  uint8_t dbgInput() {
    /* this catches input from the serial console and hands it over to inputButtons() if DEBUG is defined
       Serial input "7" matches buttonA, "8" matches buttonB, "9" matches buttonA + buttonB */
    if ( Serial.available() > 0 ) {
      uint8_t incomingByte = 0;
      incomingByte = Serial.read();
      if ( incomingByte == 52 ) {          // 4 - long press buttonA
        btnRepeatCounter = 10;
        return 1;
      }
      if ( incomingByte == 53 ) {          // 5 - long press buttonB
        btnRepeatCounter = 10;
        return 2;
      }
      if ( incomingByte == 54 ) {          // 6 - long press buttonA + buttonB
        btnRepeatCounter = 10;
        return 3;
      }
      if ( incomingByte == 55 ) return 1;  // 7 - buttonA
      if ( incomingByte == 56 ) return 2;  // 8 - buttonB
      if ( incomingByte == 57 ) return 3;  // 9 - buttonA + buttonB
    }
    return 0;
  }
 #endif
 // ---
